/******************************************************************
** File: cyucon.c
** Description: the main file for CyUCon - Cybiko Unix Console
**
**  by Jeff Frohwein - http://www.devrs.com/cybiko
** All rights reserved.
****************************************************************************
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details at www.gnu.org
****************************************************************************
** v1.0101 - Original release
** v 0.2.0 - Changes by Charles Shapiro
****************************************************************************/
#define MAIN
#include "cyucon.h"
#include "crctab.h"
#include <stdio.h> /* bufsize, perror */
#include <ctype.h> /* isspace */
/* #include <unistd.h> getopt */
#include <getopt.h> /* getopt_long */
#define VERSION "0.2.0"

enum commands { NONE,COMMAND,SEND,LISTEN,GET };

/* serial device name */
static char Device[80] = {"/dev/cybiko"};

/*
 * Command line options 
 */
static struct option Long_options[] = {
    {"command",required_argument,0,'c'},
    {"device",required_argument,0,'d'},
    {"listen",no_argument,0,'L'},
    {"memorycard",no_argument,0,'m'},
    {"manglenot",no_argument,0,'M'},
    {"receive",required_argument,0,'r'},
    {"send",required_argument,0,'s'},
    {"verbosity",required_argument,0,'v'},
    {0,0,0,0}
};



/*
Usage message
 */
void usage(int exitcode, char *program_name)
  {
  fprintf(stderr,"Usage: %s -d device -v verbosity { -m } { -M } { -s file | | -r file | -c command | -L  } \n",program_name);
  fprintf(stderr,"\nWhere:\n");
  fprintf(stderr,"\n-c sends command to the cybiko");
  fprintf(stderr,"-d device sets the serial port on which you wish ");
  fprintf(stderr,"\nto connect to your cybiko (defaults to %s)",Device);
  fprintf(stderr,"\n-L listens to the cybiko (e.g. when it boots)");
  fprintf(stderr,"\n-m causes file commands (-r and -s) to go to card flash rather than device flash");
  fprintf(stderr,"\n-M turns off filename mangling from unix to cybiko format");
  fprintf(stderr,"\n-r receives file from the cybiko");
  fprintf(stderr,"\n-s sends file to the cybiko");
  fprintf(stderr,"\n-v verbosity: higher numbers are more chatter;max is 3, default is 1");
  fprintf(stderr,"\n\nJeff Frohwein (hard part), Charles Shapiro (easy part)");
  fprintf(stderr,"\nVersion %s\n",VERSION);
}

/*
  Main line.
 */
int main(int argc, char *argv[]) 
{
  struct  termios pts;  /* termios settings on port */
  struct  termios sts;  /* termios settings on stdout/in */
  struct sigaction sact;/* used to initialize the signal handler */
  int getopt_letter;
  int option_index;
  enum commands cmd = NONE;

  /* Set defaults on arguments. */
  Verbosity=PROGRESS;
  MemoryFlash=0;
  MangleFilename=1;

  setvbuf(stdout,NULL,_IONBF,0); /* Turn buffering off on stdout.*/

  /* parse command line */
  while(-1 != (getopt_letter = getopt_long(argc,argv,"d:s:r:c:v:mML",
  Long_options,&option_index)))
  {
      switch(getopt_letter) 
      {
      case 'd':
	  memset(Device,0,sizeof(Device));
	  strncpy(Device,optarg,sizeof(Device)-1);
	  break;
      case 'm':
	  MemoryFlash=1;
	  break;
      case 'M':
	  MangleFilename=0;
	  break;
      case 'r':
	  memset(param_name,0,sizeof(param_name));
	  strncpy(param_name,optarg,sizeof(param_name)-1);
	  removeLeadingSpaces(param_name);
	  cmd=GET;
	  break;
      case 's':
	  memset(param_name,0,sizeof(param_name));
	  strncpy(param_name,optarg,sizeof(param_name)-1);
	  removeLeadingSpaces(param_name);
	  cmd=SEND;
	  break;
      case 'c':
	  memset(param_name,0,sizeof(param_name));
	  strncpy(param_name,optarg,sizeof(param_name)-1);
	  cmd=COMMAND;
	  break;
      case 'C':
	  /* Do some stuff to set up interactive mode */
	  fprintf(stderr,
	  "\nThe option you have selected is not yet implemented\n");
	  exit(1);
      case 'L':
	  /* Do some stuff to set up listen mode */
	  fprintf(stderr,"\nEntering Listen mode. Press a key to quit.");
	  cmd=LISTEN;
	  break;
      case 'v':
	  Verbosity=atoi(optarg);
	  break;
      default:
      /* Internal Error */
      }
  }

  /* open the device */
  MaybePrintMessage(PROGRESS,"Opening %s for cybiko",Device);

  /* pf = open(Device, O_RDWR|O_SYNC); */
  pf = open(Device, O_RDWR);
  if (pf < 0)
    {    
    MaybePrintMessage(ERROR,
    "Cannot open device %.80s:error %d",Device,errno);
    exit(2);
    }

  /* modify the port configuration */
  tcgetattr(pf, &pts);
  memcpy(&pots, &pts, sizeof(pots));
  init_comm(&pts);
  tcsetattr(pf, TCSANOW, &pts);
     
  /* Now deal with the local terminal side */
  tcgetattr(STDIN_FILENO, &sts);
  memcpy(&sots, &sts, sizeof(sots)); /* to be used upon exit */
  init_stdin(&sts);
  tcsetattr(STDIN_FILENO, TCSANOW, &sts);

  /* set the signal handler to restore the old
   * termios handler */
  sact.sa_handler = cleanup_termios; 
  sigaction(SIGHUP, &sact, NULL);
  sigaction(SIGINT, &sact, NULL);
  sigaction(SIGPIPE, &sact, NULL);
  sigaction(SIGTERM, &sact, NULL);

   /* run the main program loop */
   switch (cmd)
     {
     case NONE:
       usage(0,argv[0]);
       break;
     case COMMAND:
       SendCommand(pf, param_name,0,READYIND,1,1);
       printf("\n"); /* Eliminate irritating space on cmd prompt */
       break;
     case SEND:
	 SendFile(pf, param_name);
	 printf("\n"); /* Eliminate irritating space on cmd prompt */
	 break;
     case LISTEN:
	 Listen(pf);
	 break;
     case GET:
	 GetFile(pf,param_name);
	 break;
     default:
	 MaybePrintMessage(ERROR,
	 "Oops. Something went bad wrong in getopt.");
	 break;
     }
  /* restore original terminal settings and exit */
  tcsetattr(pf, TCSANOW, &pots);
  tcsetattr(STDIN_FILENO, TCSANOW, &sots);

  exit(0);

}
