/*
 * Receive a file from the cybiko. Charles Shapiro
 *
**** All rights reserved.
****************************************************************************
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details at www.gnu.org
****************************************************************************
*/

#include "cyucon.h"
#include "crctab.h"

#include <stdio.h> /* bufsize, perror */
#include <sys/stat.h> /* struct stat */
#include <sys/time.h> /* struct timeval */


#define CMDPACKETSIZE 12 /* Length of a command packet + 4-byte CRC */
#define DATAPACKETSIZE 516 /* Max length of a data multipacket (type 01) */
#define DATAPACKETHEADERSIZE 4 /* Size of a data multipacket header */
#define CRCSIZE 4 /* Size in bytes of a CRC32 */

/* 
 * Hold elements of a command packet */
typedef struct cmdPkt {
  char packetType;
  char packetCommand;
  char packetCommandTwost; /* For end packet. */
  unsigned int packetSeq;
  int packetSeqRequested;
  unsigned int packetSendLen;
  unsigned char packetSendChars[8];
  char packetCRC[CRCSIZE];
} commandPacket;

/*
Set number of characters we wish to read before returning from 
a blocked read.
 */
static void setNumCharsTimeout(int pf,unsigned char numChars,
unsigned char timeOutTenths)
{
    struct termios myTermios;

    tcgetattr(pf,&myTermios);
    myTermios.c_cc[VTIME]=timeOutTenths;
    myTermios.c_cc[VMIN]=numChars;
    tcsetattr(pf,TCSANOW,&myTermios);
}

/*
 * Create 32-bit CRC for packets exchanged on receiving a file from the
 * cybiko. This is different from the CRC used on files sent to the cybiko.  
 */
static void getFileCRC32(const char *inPacket,size_t pktLength,char *outCRC)
{
  size_t pktCtr=pktLength;
  const char *pktPtr=inPacket;
  unsigned long crcNumber=0xffffffff;

  while(pktCtr) 
  {
    crcNumber=((crcNumber >> 8) & 0x00ffffff) ^ 
      CRCTable[ (crcNumber ^ (*pktPtr)) & 0xff];
    --pktCtr;
    ++pktPtr;
  }
  crcNumber ^= 0xffffffff;
  MaybePrintMessage(DEBUG,"SendCRC is %lX");
 
  *outCRC =  (crcNumber >> 24) & 0xff;
  outCRC[1] = (crcNumber >> 16) & 0xff;
  outCRC[2] = (crcNumber >> 8) & 0xff;
  outCRC[3] = crcNumber & 0xff;
  

  MaybePrintMessage(DEBUG,"outCRC is: %hX %hX %hX %hX",
		    outCRC[0],outCRC[1],outCRC[2],outCRC[3]);
}

/*
 * Is the CRC transmitted the same as the one we calculate on the 
 * received packet?
 */
static int checkReceivedCRC(const char *inPacket,size_t packetLen)
{
  char myCRC[CRCSIZE];
  const char *sentPacketCRC;

  int status=STAT_SUCCESS;

  getFileCRC32(inPacket,packetLen-CRCSIZE,myCRC);
  sentPacketCRC=(inPacket + packetLen)-4;

  MaybePrintMessage(DEBUG,"sentCRC: %hX %hX %hX %hX", 
  sentPacketCRC[0],sentPacketCRC[1],sentPacketCRC[2],sentPacketCRC[3]);

  if(0 != memcmp(myCRC,sentPacketCRC,CRCSIZE))
    status=STAT_FAILURE;

  return status;
}
/*
Write the data in a known good packet to outfile.
 */
int writePacketToFile(const char *packetBuf,size_t packetBufLen,FILE *outfile)
{
  size_t writeSize;
  writeSize = (size_t)packetBuf[2];
  writeSize = (writeSize << 8) & 0x0000ff00;
  MaybePrintMessage(DEBUG,"writeSize after shift=%ld,packetBuf=%d",writeSize,
  (unsigned char)packetBuf[3]);
  writeSize += (unsigned char)packetBuf[3];
  MaybePrintMessage(DEBUG,"Writing %ld bytes of data",writeSize);
  packetBuf += 4;
  fwrite(packetBuf,writeSize,1,outfile);
  return STAT_SUCCESS;

}
/*
 * Figure out if we got a good packet, write it to the outfile if so.
 * If not, return failure to force resend 
 */
static int processDataPacket(const void *packetBuffer,size_t packetBufferLen,
int expectedPacketNumber, int packetDataLen, FILE *outFile)
{  
  int status=STAT_SUCCESS;


  if(packetDataLen)
  {
    status=checkReceivedCRC(packetBuffer,packetBufferLen);
    if(STAT_SUCCESS == status)
    {
      int currentPacketSequenceNum = (int)((char *)packetBuffer)[1];
      if(currentPacketSequenceNum != expectedPacketNumber)
      {
	MaybePrintMessage(ERROR,"Misnumbered packet: %d instead of %d",
			  currentPacketSequenceNum,expectedPacketNumber);
	  status=STAT_FAILURE;
      }
    }
    if(STAT_SUCCESS == status)
    {
      MaybePrintMessage(PROGRESS,"Writing packet %d",expectedPacketNumber);
      writePacketToFile(packetBuffer,packetDataLen,outFile);
    }
  }
  else
      status=STAT_NEUTRAL;
  
  return status;
}

/*
 * Send a file upload command packet to the cybiko.
 */
static int sendCmdPacket(int pf,commandPacket *inbuf)
{
  write(pf,inbuf->packetSendChars,inbuf->packetSendLen);
  write(pf,inbuf->packetCRC,sizeof(inbuf->packetCRC));

  return STAT_SUCCESS;
}

/*
 * Perform an actual data packet read
 */
int readDataPacket(int pf,char *packetBuffer,size_t numBytesToRead,
size_t *outNumRead)
{
  size_t numReadTotal=0;
  size_t numReadThisPass=0;
  int status=STAT_SUCCESS;

  while((numReadTotal < numBytesToRead) && (STAT_SUCCESS == status))
  {
    numReadThisPass = read(pf,packetBuffer,numBytesToRead);
    numReadTotal += numReadThisPass;
    packetBuffer += numReadThisPass;
    MaybePrintMessage(DEBUG,"numReadThisPass: %d of %d in %d",numReadThisPass,
		      numReadTotal,numBytesToRead);
    if(numReadThisPass <= 0)
    {
      MaybePrintMessage(DEBUG,"Read overrun: %d",numReadThisPass);
      status=STAT_NEUTRAL;
    }
  }
  if(STAT_NEUTRAL == status)
    status=STAT_SUCCESS;
  *outNumRead=numReadTotal;
  return status;
}
/*
 * Read one and only one data packet off comm line.
 */
int gatherPacket(int pf, char *packetBuffer,size_t packetBufferSize,
size_t *packetSize,int *packetDataSize)
{
    int numRead;
    int status=STAT_SUCCESS;
    size_t numDataBytes;

    status = readDataPacket(pf,packetBuffer,packetBufferSize,&numRead);
    
    if(STAT_SUCCESS == status)
    {
	*packetSize=numRead;
	numDataBytes=(((unsigned char)packetBuffer[2]) << 8) & 0x0000ff00;
	numDataBytes += (unsigned char)packetBuffer[3];
	*packetDataSize = numDataBytes; /* Size of actual data portion. */
	MaybePrintMessage(DEBUG,"Read %ld bytes of data",numDataBytes);
	if(numDataBytes > numRead)
	{
	  MaybePrintMessage(ERROR,
	  "Short or corrupted packet: read %ld of %ld",numRead,numDataBytes);
	  status=STAT_FAILURE;
	}
    }
    return status;
}

/*
 * Create a command packet to send to the cybiko
 */
static void createSendPacket(commandPacket * inOutPtr)
{
   char *packetOutput=inOutPtr->packetSendChars;
   
   memset(packetOutput,0,sizeof(inOutPtr->packetSendChars));
   *packetOutput=inOutPtr->packetType;
   packetOutput[1]=(char)inOutPtr->packetSeq;
   /* 3rd byte of packet is for packet size; always 0 */
   packetOutput[3]=(char)4; /* sent packets are always 4 bytes.*/
   packetOutput[5]=inOutPtr->packetCommand;
   packetOutput[7]=inOutPtr->packetCommandTwost; /* only for "quit" packet. */
   if(inOutPtr->packetSeqRequested > -1)
   {
     packetOutput[6]=inOutPtr->packetSeqRequested >> 8;
     packetOutput[7]=inOutPtr->packetSeqRequested & 0xff;
   }
   inOutPtr->packetSendLen=sizeof(inOutPtr->packetSendChars); /* so far */

   MaybePrintMessage(DEBUG,
    "PacketSendChars: %hX %hX %hX %hX %hX %hX %hX %hX",
    inOutPtr->packetSendChars[0], inOutPtr->packetSendChars[1],
    inOutPtr->packetSendChars[2], inOutPtr->packetSendChars[3],
    inOutPtr->packetSendChars[4], inOutPtr->packetSendChars[5],
    inOutPtr->packetSendChars[6], inOutPtr->packetSendChars[7]);
 
   getFileCRC32(inOutPtr->packetSendChars,inOutPtr->packetSendLen,
		inOutPtr->packetCRC);
   return;
}

/*
Send and receive first pair of packets; always the same.
*/
static int initializePacketExchange(int pf)
{
  commandPacket sendPacket;
  char myPacketBuffer[CMDPACKETSIZE+1];
  size_t rcvedLen;
  int status=STAT_SUCCESS;
 
  memset(&sendPacket,0,sizeof(sendPacket));
  sendPacket.packetSeq=0;
  sendPacket.packetType=(char)2;
  sendPacket.packetCommand=(char)5;
  sendPacket.packetSeqRequested=-1; /* An initialization packet */
  createSendPacket(&sendPacket);
  setNumCharsTimeout(pf,0,7);

  sendCmdPacket(pf,&sendPacket);
  readDataPacket(pf,myPacketBuffer,sizeof(myPacketBuffer)-1,&rcvedLen);

  if (myPacketBuffer[5] != (char)1) {
    MaybePrintMessage(ERROR,
    "Ouch. This is not a data packet: %s",myPacketBuffer);
    status=STAT_FAILURE;
  }

  if( (STAT_SUCCESS == status) && (CMDPACKETSIZE != rcvedLen) )
    {
      MaybePrintMessage(ERROR,"Read wrong number of bytes: %d out of %d",
			rcvedLen,CMDPACKETSIZE);
      status=STAT_FAILURE;
    }
  
  if(STAT_SUCCESS == status)
  {
    if(STAT_SUCCESS != checkReceivedCRC(myPacketBuffer,rcvedLen))
    {
      MaybePrintMessage(ERROR,"CRC check failure in initialization");
      status=STAT_FAILURE;
    }    
  }
  return status;
}

/*
Get a file from the Cybiko.
 */
void GetFile(int pf, char *fileName)
{
    int finished=0;
    int rcvedPacketSequence=0;
    int gotPacket;
    FILE *outFile;
    char cmdBuffer[80];
    char fileReceiveBuffer[2048];
    unsigned int packetDataLen;
    commandPacket sendPacket;
    int processStatus=STAT_SUCCESS;
    size_t packetReceivedLen=0;
    int status=STAT_SUCCESS;
    char *insideFileName=calloc(1,strlen(fileName)+5);
    char finishedIndicator[5];

    if(MangleFilename)
      fixFileNameOutOfCybiko(fileName,insideFileName);
    else
      strcpy(insideFileName,fileName);
    if(MemoryFlash)
    {
      char *tempBuf=calloc(1,strlen(insideFileName)+5);
      fileNameToMemCard(insideFileName,tempBuf,strlen(insideFileName)+5);
      strcpy(insideFileName,tempBuf);
      free(tempBuf);
			   
    }
    outFile = fopen(fileName,"w");
    setvbuf(outFile,NULL,_IONBF,0); /* Buffering off on outfile for debug.*/
    memset(cmdBuffer,0,sizeof(cmdBuffer));
    strcpy(cmdBuffer,"upl ");
    strncat(cmdBuffer,insideFileName,
	    sizeof(cmdBuffer) - (strlen(cmdBuffer)+1));
    memset(finishedIndicator,0,sizeof(finishedIndicator));
    *finishedIndicator=cmdBuffer[strlen(cmdBuffer)-1];
    SendCommand(pf,"pc on",0,READYIND,1,1);
    SendCommand(pf,cmdBuffer,0,finishedIndicator,1,1);
    /*    delay(0,500);*/
    if(STAT_SUCCESS == status)
      status = initializePacketExchange(pf);

    while( (! finished)  && (STAT_SUCCESS == status) )
    {
      sendPacket.packetSeq=0;
      sendPacket.packetType=(char)2;
      sendPacket.packetCommand=(char)3;
      sendPacket.packetSeqRequested=rcvedPacketSequence;
      createSendPacket(&sendPacket);
      sendCmdPacket(pf,&sendPacket);	   
      
      status = gatherPacket(pf,fileReceiveBuffer,sizeof(fileReceiveBuffer),
      &packetReceivedLen,&packetDataLen);

      processStatus = processDataPacket(fileReceiveBuffer,packetReceivedLen,
      rcvedPacketSequence,packetDataLen,outFile);

      switch(processStatus) {
      case STAT_SUCCESS:
	  gotPacket=0;
	  packetReceivedLen=0;
	  ++rcvedPacketSequence;	
	  break;
      case STAT_NEUTRAL: /* End of transmission. */
	  finished=1;
	  break;
      default: /* STAT_FAILURE: re-ask for same packet. */
	  packetReceivedLen=0;
	  break;
      } /* switch(processStatus) */
    } /* while((! finished) && (STAT_SUCCESS == status)) */

    if(STAT_SUCCESS == status)
    { 
     
      sendPacket.packetSeq=0;
      sendPacket.packetType=(char)2;
      sendPacket.packetCommand=(char)2;
      sendPacket.packetCommandTwost=(char)2;
      sendPacket.packetSeqRequested=rcvedPacketSequence;
      createSendPacket(&sendPacket);
      sendCmdPacket(pf,&sendPacket);	   	
    }
    SendCommand(pf,"pc off",0,READYIND,1,1);
    fclose(outFile);
    free(insideFileName);
}




